/*
 * Copyright (C)2003 USAGI/WIDE Project
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * Authors:
 *	Noriaki TAKAMIYA @USAGI
 *	Masahide NAKAMURA @USAGI
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/skbuff.h>
#include <linux/ipv6.h>
#include <net/ipv6.h>
#include <net/mip6.h>
#include <net/protocol.h>
#include <net/xfrm.h>


unsigned inline int calc_padlen(unsigned int len)
{
	return (8 - (len % 8));
}

static inline void mip6_padn(__u8 *data, __u8 padlen)
{
	if (padlen < 1 || !data)
		return;

	if (padlen == 1)
		data[0] = MIP6_OPT_PAD_1;
	else {
		data[0] = MIP6_OPT_PAD_N;
		data[1] = padlen - 2;
		if (padlen > 2)
			memset(data+2, 0,  data[1]);
	}
}

int mip6_dest_hao(struct sk_buff *skb, int optoff)
{
	struct in6_addr *haddr;

	if (skb->nh.raw[optoff+1] != 16) {
#ifdef CONFIG_IPV6_MIP6_DEBUG
		printk(KERN_DEBUG "%s: invalid option length of haddr: %d\n", __FUNCTION__,
			skb->nh.raw[optoff+1]);
#endif
		kfree_skb(skb);
		return 0;
	}

	haddr = (struct in6_addr *)(skb->nh.raw+optoff+2);

	if (!(ipv6_addr_type(haddr) & IPV6_ADDR_UNICAST)) {
#ifdef CONFIG_IPV6_MIP6_DEBUG
		printk(KERN_DEBUG "%s: invalid haddr: %x:%x:%x:%x:%x:%x:%x:%x\n", __FUNCTION__,
			NIP6(*haddr));
#endif
		kfree_skb(skb);
		return 0;
	}

#ifdef CONFIG_IPV6_MIP6_DEBUG
	printk(KERN_DEBUG "%s: haddr: %x:%x:%x:%x:%x:%x:%x:%x\n", __FUNCTION__,
		NIP6(*haddr));
#endif

	((struct inet6_skb_parm*)skb->cb)->hao = optoff+2;
	return 1;
}

static int mip6_rcv(struct sk_buff **skbp, unsigned int *nhoffp)
{
	struct sk_buff *skb = *skbp;
	struct mip6_mh_hdr *mh;

#ifdef CONFIG_IPV6_MIP6_DEBUG
	printk(KERN_DEBUG "%s: called\n", __FUNCTION__);
#endif

	if (!pskb_may_pull(skb, (skb->h.raw - skb->data) + 8) ||
		!pskb_may_pull(skb, (skb->h.raw - skb->data) + ((skb->h.raw[1] + 1) << 3))) {
		kfree_skb(skb);
		return -1;
	}

	/* operation for mobility header */
	mh = (struct mip6_mh_hdr *)skb->h.raw;

	/* XXX: verify checksum; we need code */
	;

	/* payload proto */
	if (mh->nexthdr != IPPROTO_NONE) {
#ifdef CONFIG_IPV6_MIP6_DEBUG
		printk(KERN_DEBUG "%s: invalid payload proto:%d\n", __FUNCTION__,
			mh->nexthdr);
#endif
		kfree_skb(skb);
		return -1;
	}

	/* next protocol process */
	skb->h.raw += ((mh->hdrlen + 1) << 3);
	*nhoffp = (&mh->nexthdr) - skb->nh.raw;

	return 1;
}

void mip6_err(struct sk_buff *skb, struct inet6_skb_parm *opt, 
	      int type, int code, int offset, __u32 info)
{
#ifdef CONFIG_IPV6_MIP6_DEBUG
	printk(KERN_DEBUG "%s: called\n", __FUNCTION__);
#endif
}

/*****************************
 * Common part
 *****************************/

int mip6_output(struct sk_buff *skb,
		int (*append)(struct sk_buff *skb, u16 offset, u8 nexthdr))
{
	struct dst_entry *dst = skb->dst;
	struct xfrm_state *x = dst->xfrm;
	struct ipv6hdr *iph = NULL;
	int hdr_len;
	__u8 *prevhdr;
	__u8 nexthdr = 0;
	int err = 0;

	spin_lock_bh(&x->lock);

	err = xfrm_check_output(x, skb, AF_INET);
	if (err) {
#ifdef CONFIG_IPV6_MIP6_DEBUG
		printk(KERN_DEBUG "%s: xfrm_check_output failed:%d (currently ignored)\n", __FUNCTION__, err);
#endif
		/*goto error;*/ /* XXX: Currently ignored this error. */
		err = 0;
	}

	if (!append)
		goto output;

	hdr_len = ip6_find_1stfragopt(skb, &prevhdr);
	nexthdr = *prevhdr;
	*prevhdr = x->id.proto; /* XXX: */

	iph = kmalloc(hdr_len, GFP_ATOMIC);
	if (!iph) {
		err = -ENOMEM;
		goto error;
	}
	memcpy(iph, skb->data, hdr_len);
	skb->nh.ipv6h = (struct ipv6hdr*)skb_push(skb, x->props.header_len);
	memcpy(skb->nh.ipv6h, iph, hdr_len);
	kfree(iph);
	iph = NULL;

	err = append(skb, hdr_len, nexthdr);
	if (err)
		goto error;

	skb->nh.ipv6h->payload_len = htons(skb->len - sizeof(struct ipv6hdr));

	skb->nh.raw = skb->data;

 output:
	x->curlft.bytes += skb->len;
	x->curlft.packets++;
	spin_unlock_bh(&x->lock);
	if ((skb->dst = dst_pop(dst)) == NULL) {
#ifdef CONFIG_IPV6_MIP6_DEBUG
		printk(KERN_DEBUG "%s: dst_pop is NULL\n", __FUNCTION__);
#endif
		err = -EHOSTUNREACH;
		goto error_nolock;
	}

#ifdef CONFIG_IPV6_MIP6_DEBUG
	printk(KERN_DEBUG "%s: NET_XMIT_BYPASS\n", __FUNCTION__);
#endif
	return NET_XMIT_BYPASS;

 error:
	spin_unlock_bh(&x->lock);
 error_nolock:
	kfree_skb(skb);
	return err;
}

/*****************************
 * Destination Options Header
 *****************************/
static int mip6_destopt_input(struct xfrm_state *x, struct xfrm_decap_state *decap, struct sk_buff *skb)
{
#ifdef CONFIG_IPV6_MIP6_DEBUG
	printk(KERN_DEBUG "%s: called\n", __FUNCTION__);
#endif
	return 0;
}

/* Destination Option Header is inserted.
 * IP Header's src address is replaced with Home Address Option in
 * Destination Option Header.
 */
static int mip6_destopt_append(struct sk_buff *skb, u16 offset, u8 nexthdr)
{
	struct dst_entry *dst = skb->dst;
	struct xfrm_state *x = dst->xfrm;
	struct ipv6hdr *iph = skb->nh.ipv6h;
	struct ipv6_destopt_hdr *dstopt;
	struct destopt_hao *hao;

	dstopt = (struct ipv6_destopt_hdr *)(skb->nh.raw + offset);
	dstopt->nexthdr = nexthdr;
	dstopt->hdrlen = 24; /* XXX: in 8-octets... */
	dstopt->hdrlen = (dstopt->hdrlen >> 3) - 1;

	hao = (struct destopt_hao *)(dstopt+1);
	hao->type = IPV6_TLV_HAO;
	hao->length = 16;

	/* XXX: 8n+6: fill padding... Fix me! */
	mip6_padn((char *)(hao+1), calc_padlen(20));

#if 1
	/* Replacing address */
	memcpy(&hao->addr, &iph->saddr, sizeof(hao->addr));
	memcpy(&iph->saddr, &x->mip6_state.addr, sizeof(iph->saddr));
#ifdef CONFIG_IPV6_MIP6_DEBUG
	printk(KERN_DEBUG "%s: replacement results:\n", __FUNCTION__);
	printk(KERN_DEBUG "  src=%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x\n", NIP6(iph->saddr));
	printk(KERN_DEBUG "  hao=%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x\n", NIP6(hao->addr));
#endif
#else
	memcpy(&hao->addr, &x->mip6_state.addr, sizeof(hao->addr));
#endif
	skb->h.raw = (unsigned char*)dstopt;

	return 0;
}

static int mip6_destopt_output(struct sk_buff *skb)
{
	BUG_TRAP(skb->dst->xfrm->id.proto == IPPROTO_DSTOPTS);
	return mip6_output(skb, mip6_destopt_append);
}

static int mip6_destopt_init_state(struct xfrm_state *x, void *args)
{
#ifdef CONFIG_IPV6_MIP6_DEBUG
	printk(KERN_DEBUG "%s: called\n", __FUNCTION__);
#endif
	x->props.header_len = 24; /* XXX: */

	return 0;
}

static void mip6_destopt_destroy(struct xfrm_state *x)
{
#ifdef CONFIG_IPV6_MIP6_DEBUG
	printk(KERN_DEBUG "%s: called\n", __FUNCTION__);
#endif
}

/*****************************
 * Routing Header type 2
 *****************************/
static int mip6_rthdr_input(struct xfrm_state *x, struct xfrm_decap_state *decap, struct sk_buff *skb)
{
#ifdef CONFIG_IPV6_MIP6_DEBUG
	printk(KERN_DEBUG "%s: called\n", __FUNCTION__);
#endif
	return 0;
}

/* Routing Header type 2 is inserted.
 * IP Header's dst address is replaced with Routing Header's Home Address.
 */
static int mip6_rthdr_append(struct sk_buff *skb, u16 offset, u8 nexthdr)
{
	struct dst_entry *dst = skb->dst;
	struct xfrm_state *x = dst->xfrm;
	struct ipv6hdr *iph = skb->nh.ipv6h;
	struct rt2_hdr *rt2;

	rt2 = (struct rt2_hdr *)(skb->nh.raw + offset);
	rt2->rt_hdr.nexthdr = nexthdr;
	rt2->rt_hdr.hdrlen = 2; /* XXX: in 8-octets... */
	rt2->rt_hdr.type = 2;
	rt2->rt_hdr.segments_left = 1;
	memset(&rt2->reserved, 0, sizeof(rt2->reserved));

#if 1
	/* Replacing address */
	memcpy(&rt2->addr, &iph->daddr, sizeof(rt2->addr));
	memcpy(&iph->daddr, &x->mip6_state.addr, sizeof(iph->daddr));
#ifdef CONFIG_IPV6_MIP6_DEBUG
	printk(KERN_DEBUG "%s: replacement results:\n", __FUNCTION__);
	printk(KERN_DEBUG "  dst=%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x\n", NIP6(iph->daddr));
	printk(KERN_DEBUG "  rt2=%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x\n", NIP6(rt2->addr));
#endif
#else
	memcpy(&rt2->addr, &x->mip6_state.addr, sizeof(rt2->addr));
#endif

	skb->h.raw = (unsigned char*)rt2;

	return 0;
}

static int mip6_rthdr_output(struct sk_buff *skb)
{
	BUG_TRAP(skb->dst->xfrm->id.proto == IPPROTO_ROUTING);
	return mip6_output(skb, mip6_rthdr_append);
}

static int mip6_rthdr_init_state(struct xfrm_state *x, void *args)
{
#ifdef CONFIG_IPV6_MIP6_DEBUG
	printk(KERN_DEBUG "%s: called\n", __FUNCTION__);
#endif
	x->props.header_len = sizeof(struct rt2_hdr);

	return 0;
}

static void mip6_rthdr_destroy(struct xfrm_state *x)
{
#ifdef CONFIG_IPV6_MIP6_DEBUG
	printk(KERN_DEBUG "%s: called\n", __FUNCTION__);
#endif
}

/*****************************
 * Module part
 *****************************/
static struct xfrm_type mip6_destopt_type =
{
	.description	= "MIP6",
	.owner		= THIS_MODULE,
	.proto	     	= IPPROTO_DSTOPTS,
	.init_state	= mip6_destopt_init_state,
	.destructor	= mip6_destopt_destroy,
	.input		= mip6_destopt_input,
	.output		= mip6_destopt_output
};

static struct xfrm_type mip6_rthdr_type =
{
	.description	= "MIP6",
	.owner		= THIS_MODULE,
	.proto	     	= IPPROTO_ROUTING,
	.init_state	= mip6_rthdr_init_state,
	.destructor	= mip6_rthdr_destroy,
	.input		= mip6_rthdr_input,
	.output		= mip6_rthdr_output
};

static struct inet6_protocol mip6_protocol = {
	.handler	=	mip6_rcv,
	.err_handler	=	mip6_err
};

int __init mip6_init(void)
{
	printk(KERN_INFO "Mobile IPv6\n");

	if (xfrm_register_type(&mip6_destopt_type, AF_INET6) < 0) {
		printk(KERN_INFO "%s: can't add xfrm type(destopt)\n", __FUNCTION__);
		goto mip6_destopt_xfrm_fail;
	}
	if (xfrm_register_type(&mip6_rthdr_type, AF_INET6) < 0) {
		printk(KERN_INFO "%s: can't add xfrm type(rthdr)\n", __FUNCTION__);
		goto mip6_rthdr_xfrm_fail;
	}
	if (inet6_add_protocol(&mip6_protocol, IPPROTO_MOBILITY) < 0) {
		printk(KERN_ERR "%s: can't add protocol\n", __FUNCTION__);
		goto proto_fail;
	}

	return 0;

 proto_fail:
	xfrm_unregister_type(&mip6_rthdr_type, AF_INET6);
 mip6_rthdr_xfrm_fail:
	xfrm_unregister_type(&mip6_destopt_type, AF_INET6);
 mip6_destopt_xfrm_fail:
	return -EAGAIN;
}

#if 0 /* XXX: Fix to be called */
static void __exit mip6_fini(void)
{
	if (inet6_del_protocol(&mip6_protocol, IPPROTO_MOBILITY) < 0)
		printk(KERN_INFO "%s: can't remove protocol\n", __FUNCTION__);
	if (xfrm_unregister_type(&mip6_rthdr_type, AF_INET6) < 0)
		printk(KERN_INFO "%s: can't remove xfrm type(rthdr)\n", __FUNCTION__);
	if (xfrm_unregister_type(&mip6_destopt_type, AF_INET6) < 0)
		printk(KERN_INFO "%s: can't remove xfrm type(destopt)\n", __FUNCTION__);
}
#endif
